from reaction import Reaction,Compound
from utils import *
class RecogData():
    def __init__(self,
                ocr_list = [],
                smiles_list = []
                ):
        self.ocr_list = ocr_list
        self.smiles_list = smiles_list
        self.ocr_positions = []
        self.smiles_positions = []
    def get(self):
        print(1)

class ChemData():
    def __init__(self, ocr, smiles):
        self._raw_data = RecogData(ocr, smiles)
        self.correct_data = None
    @property
    def raw_data(self):
        return self._raw_data
    def process_raw_data(self, ):
        self.possible_enzymes = []
        self.possible_compounds = []
        for e in self.raw_data.ocr_list:
            self.possible_enzymes += ChemUtils.dissolve_enzyme_synonym(e)
        for e in self.raw_data.ocr_list:
            self.possible_compounds += [Compound(input = e, init_mode='name')]
        for e in self.raw_data.smiles_list:
            self.possible_compounds += [Compound(input = e, init_mode='smiles')]
    def predict_reactions(self,):
        self.only_enzyme = [ChemUtils.find_reaction(x,) for x in self.possible_enzymes]
        self.only_cid = [ChemUtils.find_reaction(ec = [], cid=[x.cid]) for x in self.possible_compounds if x.pcp_valid]
        self.only_enzyme_reaction = [Reaction(data = ChemUtils.get_brenda_reaction(id[0][0])) for id in self.only_enzyme if id]
        self.only_cid_reaction = [Reaction(data = ChemUtils.get_brenda_reaction(id[0][0])) for id in self.only_cid if id]
    def show_sim(self):
        for j in self.only_enzyme_reaction:
            j.similarities(compounds=self.possible_compounds, enzymes=self.possible_enzymes)
            print(j.sim_compounds)
        for j in self.only_cid_reaction:
            j.similarities(compounds=self.possible_compounds, enzymes=self.possible_enzymes)
            print(j.sim_compounds)
    
if __name__ == "__main__":
    a = ChemData(
        ['Glycerol', 'NADP', 'GlyDH', 'Glyceraldehyde'],
        ['C(C(CO)O)O']
    )
    a.process_raw_data()
    # print(a.possible_enzymes)
    a.predict_reactions()
    print(a.only_enzyme[0][0])
    print(ChemUtils.get_brenda_reaction(265)['cems'])
    print(a.only_cid_reaction)
    a.show_sim()


import pubchempy as pcp
import pickle
import json
from tqdm import tqdm
from rdkit import DataStructs, Chem
from utils import ChemUtils
class Enzyme():
    def __init__(
        self,
        standard_name = '',
        name =  '',
    ):
        self.standard_name = standard_name
        self.name = name
        self.synonyms = ChemUtils.get_syns(standard_name) if self.standard_name else []
    def calc_similarity(self, other):
        ans = 0
        for syn in other.synonyms:
            ans = max(ans, ChemUtils.str_sim(syn, self.name))
        return ans
class Compound():
    def __init__(
        self,
        input,
        init_mode
    ):
        self.name = input if init_mode == 'name' else None
        try:
            self.pcp_data = pcp.get_compounds(input.lower(), init_mode)[0]
            self.cid = self.pcp_data.cid
            self.compound = pcp.Compound.from_cid(self.cid)
            self.smiles = self.pcp_data.isomeric_smiles
            self.name = self.pcp_data.iupac_name
            self.pcp_valid = True
        except Exception:
            self.pcp_valid = False
            print(f'no found pcp for {input}',)
        
        self.rd_valid = True
        if self.pcp_valid:
            self.rd_data = Chem.MolFromSmiles(self.smiles)
        elif init_mode == 'smiles':
            self.rd_data = Chem.MolFromSmiles(input)
        else:
            self.rd_valid = False
            
        
    def calc_similarity(self, other):
        if self.rd_valid and other.rd_valid:
            ms = [self.rd_data, other.rd_data]
            fps = [Chem.RDKFingerprint(x) for x in ms]
            return DataStructs.FingerprintSimilarity(fps[0],fps[1])
        elif self.name:
            return ChemUtils.str_sim(self.name, other.name)
        else:
            return 0
class Reaction():
    def __init__(self, substrate = None, products = None, enzyme = None, data = {}):
        self._enzyme = enzyme if enzyme else Enzyme()
        self.substrate = substrate
        self.products = products
        if data:
            self._enzyme = Enzyme(standard_name = data['ec_name'])
            self.compounds = [Compound(input = x, init_mode='name') for x in data['cems']]
    @property
    def enzyme(self):
        return self._enzyme
    def similarities(self, compounds, enzymes):
        """this is the place for finding similarities

        Args:
            compounds (list): the possible compounds list
            enzymes (list): the possible enzymes
        """
        tot_cnt = 0
        sim_com = 0
        for com in self.compounds:
            tot_cnt += 1
            for c in compounds:
                sim_com += com.calc_similarity(c)
        sim_com /= tot_cnt
        self.sim_compounds = sim_com


import pubchempy as pcp
import pickle
import json
from tqdm import tqdm
from rdkit import DataStructs, Chem
from nltk.metrics import *
def load_json(path):
    with open(path, 'r', encoding='utf-8') as f:
        data = json.load(f)
    return data

class ChemUtils():
    @staticmethod
    def dynamic_leven_dist(str1, str2):
        result = edit_distance_align(str1, str2, substitution_cost=1)
        return result
    @staticmethod
    def leven_dist(str1, str2):
        return edit_distance(str1, str2, substitution_cost=1)
    @staticmethod
    def search_compounds_list(mols):
        ans = []
        for m in mols:
            ans += pcp.get_compounds(m, 'name')
        return ans
    @staticmethod
    def str_sim(str1, str2):
        tot_len = len(str1)
        dist = ChemUtils.leven_dist(str1, str2)
        good = tot_len - dist
        if(good <= 0):
            good = 0
            return good/tot_len
        else:
            return good/tot_len
    @classmethod
    def load_brenda(
        cls, 
        brenda_datapath = 'data/BrendaIDwithCid_duplicated.pkl', 
        syn_datapath = ''
        ):
        with open(brenda_datapath, 'rb') as handle:
            cls.brenda = pickle.load(handle)
        print("brenda length:", len(cls.brenda))
    @classmethod
    def init_syns(cls):    # load all synonyms of ec
        '''
        syns_list: all synonyms of ec
        reverse_dict: map synonyms to standard names
        '''
        syn_file = 'data/synonyms.json'
        syns = load_json(syn_file)
        syns_list = []
        for key in syns:
            syns[key] += [key]
            syns[key] = list(set(syns[key]))
            syns_list += syns[key]
        print(len(syns_list))

        reverse_dict = {}
        for key in syns:
            for item in syns[key]:
                if item.lower() in reverse_dict:
                    reverse_dict[item.lower()].append(key.lower())
                else:
                    reverse_dict[item.lower()] = [key.lower()]
        print(len(reverse_dict))
        cls.reverse_dict = reverse_dict
        cls.dict = syns
        return syns, syns_list, reverse_dict
    @classmethod
    def get_syns(cls, enzyme):
        return cls.dict[enzyme]
    @classmethod
    def dissolve_enzyme_synonym(cls, name):
        try:
            return cls.reverse_dict[name.lower()]
        except Exception:
            print(f'no enzyme found for {name}')
            return []
    @classmethod
    def find_reactions(cls, duplicate_pairs, ):
        brenda = cls.brenda
        reaction_pairs = []
        for item in duplicate_pairs:
            ec = item[0]
            cid = item[1]
            # print(ec, cid)
            reaction_ids = cls.find_reactions(ec, cid)
            if reaction_ids != []:
                reaction_pairs.append([ec, cid, reaction_ids])
        print("reaction pairs:", len(reaction_pairs))
        return reaction_pairs
    @classmethod
    def find_reaction(cls, ec = [], cid = []):
        """find reaction according to ec and cids

        Args:
            ec (_type_): the enzyme standardized name
            cid (_type_): the cid
        """
        brenda = cls.brenda
        # print(ec)
        reaction_ids = [] # save mapped react id and ent name
        if ec != []:
            for key in brenda:
                if brenda[key]['ec_name'].lower() == ec:
                    set_ex_cid = set(cid)
                    for idx, cand_cids in enumerate(brenda[key]['cids']):
                        set_cand = set(cand_cids)
                        # print(set_cand)
                        if (set_cand & set_ex_cid) or cid == []:
                            cand_ent_name = brenda[key]['cems'][idx] # save the mapped ent name
                            reaction_ids.append([key, cand_ent_name])
        else:
            print(cid)
            set_ex_cid = set(cid)
            for key in brenda:
                for idx, cand_cids in enumerate(brenda[key]['cids']):
                    set_cand = set(cand_cids)
                    # print(set_cand)
                    if (set_cand & set_ex_cid) or cid == []:
                        cand_ent_name = brenda[key]['cems'][idx] # save the mapped ent name
                        reaction_ids.append([key, cand_ent_name])
        return reaction_ids
    @classmethod
    def get_brenda_reaction(cls, id):
        return cls.brenda[id]
ChemUtils.load_brenda()
ChemUtils.init_syns()

if __name__ == "__main__":
    ChemUtils.load_brenda()
    ChemUtils.init_syns()
    a = ChemUtils.dissolve_enzyme_synonym('GlyDH')
    print(a)
    print(ChemUtils.leven_dist('gly', '-glyfdas8'))
    print(ChemUtils.str_sim('gly', '-gly'))